# Copyright (c) 2023 - 2025, AG2ai, Inc., AG2ai open-source projects maintainers and core contributors
#
# SPDX-License-Identifier: Apache-2.0

# Prefix for all wrapped agent names
__AGENT_WRAPPER_PREFIX__ = "wrapped_"
